#include <vector>
#include <cassert>

#include "constraints.h"
#include "graph.h"

bool NotAllNeighboursConstraint::check(Graph &G) {
	size_t count = 0;
	Vertex *maybe_should_be_permanent = nullptr;

	for (Vertex *a: neighbours) {
		Vertex *m = a;
		while (m->merge != nullptr) {
			assert(m->cstr_in_sol == false);
			m = m->merge;
		}
		if (m->cstr_in_sol) {
			count++;
		}
		else {
			maybe_should_be_permanent = m;
		}
	}

	if (neighbours.size() - count == 1) {
		mark_permanent(G, maybe_should_be_permanent);
	}

	return count != neighbours.size();
}

NotAllNeighboursConstraint::NotAllNeighboursConstraint(Vertex *V) : v(V) {
	for (auto edge: v->edges) {
		neighbours.push_back(edge.first);
	}
}

NoPermanentVertexConstraint::NoPermanentVertexConstraint() {
}

bool NoPermanentVertexConstraint::check(Graph &G) {
	for (Vertex *v : G.permanent) {
		if (v->merge != nullptr) {
			assert(v->cstr_in_sol == false);
			Vertex *m = v;
			while (m->merge != nullptr)
				m = m->merge;

			if (m->cstr_in_sol)
				return false;
		}
		if (v->cstr_in_sol)
			return false;
	}

	return true;
}
