#include <cassert>

#include "graph.h"
#include "optimize.h"
#include "snapshot.h"

bool tc_simple_choice(Graph &G, long long &k) {
	long long orig_k = k;

	vector<Vertex *> V_copy(G.V);
	for (Vertex *a: V_copy) {
		vector<Vertex *> marked;
		vector<Vertex *> current;
		vector<Vertex *> next;

		assert(!a->marked);
		a->marked = true;
		a->dist   = 0;
		a->tmp    = nullptr;
		marked.push_back(a);
		current.push_back(a);


		// BFS
		for (int dist = 0; dist < 3; dist++) {
			for (Vertex *b: current) {
				assert(b->dist == dist);
				assert(b->marked);

				b->score = b->permanent ? 1000000ll : b->weight;
				if (dist != 0)
					b->score += b->tmp->score;

				for (auto edge: b->edges) {
					Vertex *c = edge.first;

					if (c->marked) {
						if (c == a)
							continue;
						assert(b->dist >= c->tmp->dist);
						// allow to set the parent of c to b if b is better than c->tmp
						if (b->dist != c->tmp->dist || b->score < c->tmp->score)
							continue;
					}

					c->tmp = b;
					c->dist = dist+1;

					if (!c->marked) {
						c->marked = true;
						next.push_back(c);
						marked.push_back(c);
					}
				
				}
			}

			current.clear();
			swap(current, next);
		}

		for (Vertex *v: current) {
			assert(v->dist == 3);
			Vertex *u = v->tmp;
			Vertex *t = u->tmp;
			Vertex *s = t->tmp;
			assert(s == a);

			if (s->deg == 0 || t->deg == 0 || u->deg == 0 || v->deg == 0)
				continue;

			long long score = 0;
			if (s->permanent)
				score++;
			if (t->permanent)
				score++;
			if (u->permanent)
				score++;
			if (v->permanent)
				score++;

			if (score == 4)
				add_to_vds(G, k, s);

			if (score == 3) {
				Vertex *rem = nullptr;
				if (!s->permanent)
					rem = s;
				if (!t->permanent)
					rem = t;
				if (!u->permanent)
					rem = u;
				if (!v->permanent)
					rem = v;

				add_to_vds(G, k, rem);
			}
		}

		// cleanup
		for (Vertex *b: marked) {
			assert(b->marked);
			b->marked = false;
			b->tmp = nullptr;
			b->dist = 0;
		}

		for (Vertex *b: G.V)
			assert(!b->marked);
		marked.clear();
		current.clear();
		next.clear();
	}

	return k != orig_k;
}

static void dfs(Vertex *v, size_t nr, vector<Vertex *> &component) {
	if (v->component != 0) {
		assert(v->component == nr);
		return;
	}
	v->component = nr;
	component.push_back(v);

	for (auto edge: v->edges) {
		Vertex *a = edge.first;
		dfs(a, nr, component);
	}
}

static bool is_2club(vector<Vertex *> component) {
	vector<Vertex *> marked;
	vector<Vertex *> current;
	vector<Vertex *> next;
	int dist;

	for (Vertex *a: component) {
		// find restricted P4 beginning with a
		bool found = false;

		dist = 1;
		marked.push_back(a);
		assert(!a->marked);
		a->marked = true;

		for (auto edge: a->edges) {
			Vertex *b = edge.first;
			b->marked = true;
			b->tmp    = a;

			current.push_back(b);
			marked.push_back(b);
		}


		// BFS
		while (dist != 3) {
			for (Vertex *b: current) {
				for (auto edge: b->edges) {
					Vertex *c = edge.first;
					if (c->marked)
						continue;

					c->marked = true;
					c->tmp    = b;

					next.push_back(c);
					marked.push_back(c);
				
				}
			}

			current.clear();
			swap(current, next);
			dist++;
		}

		if (current.size() != 0) {
			found = true;
		}

		// cleanup
		for (Vertex *b: marked) {
			assert(b->marked);
			b->marked = false;
			b->tmp = nullptr;
		}
		marked.clear();
		current.clear();
		next.clear();

		if (found)
			return false;
	}

	return true;
}

bool tc_2club_purge(Graph &G) {
	bool success = false;
	vector<Vertex *> V_copy(G.V);
	for (Vertex *v: V_copy) {
		v->component = 0;
		assert(v->component == 0);
	}

	size_t comp_nr = 1;
	for (Vertex *v: V_copy) {
		if (v->component != 0)
			continue;

		vector<Vertex *> component;
		dfs(v, comp_nr, component);
		comp_nr++;

		if (is_2club(component)) {
			success = true;
			for (Vertex *a: component) {
				delete_vertex(G, a);
			}
		}
	}

	for (Vertex *v: V_copy) {
		v->component = 0;
	}
	
	return success;
}
